/*

*/

package language3;

import tree.*;
import tokenizer.*;

/**
   Evaluate (interpret) an Abstract Syntax Tree (AST)
   that is defined by the following grammar.
   <pre>{@code
       Prog ::= Expr
              | '(' 'prog' Expr+ ')'

       Expr ::= Begin
              | Set
              | Var
              | Print
              | BExp
              | AExp
              | INTEGER
              | BOOLEAN
              | VARIABLE

      Begin ::= '(' 'begin' Expr+ ')'

        Set ::= '(' 'set' VARIABLE Expr ')'

        Var ::= '(' 'var' VARIABLE Expr ')'

      Print ::= '(' 'print' Expr ')'

      BExpr ::= '(' '||'  Expr Expr+ ')'
              | '(' '&&'  Expr Expr+ ')'
              | '(' '!'   Expr ')'
              | '('  EqOp Expr Expr ')'
              | '(' RelOp Expr Expr ')'

       EqOp ::= '==' | '!='
      RelOp ::= '<' | '>' | '<=' | '>='

      AExpr ::= '(' '+' Expr Expr* ')'
              | '(' '-' Expr Expr? ')'
              | '(' '*' Expr Expr+ ')'
              | '(' '/' Expr Expr  ')'
              | '(' '%' Expr Expr  ')'
              | '(' '**' Expr Expr ')'

    INTEGER ::= [-|+][0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
   }</pre>
*/
public class EvaluateAST_no_error_checking
{
   public static int DEBUG = 1;

   /**
      The methods evaluateProg(), evaluateExpr(), evaluateAexpr(), evaluateBexpr(),
      and evaluateRexpr() are essentially a post-order traversal of the abstract
      syntax tree.

      @param tree  an expression {@link Tree} to be evaluated
      @return the {@link Value} of the expression {@link Tree}
   */
   public static Value eval(final Tree<Token> tree)
   {
      // Instantiate a global environment object.
      final Environment env = new Environment();

      return evaluateProg( tree, env );
   }


   /**
      Evaluate a program using the given {@link Environment}.

      @param tree  an expression {@link Tree} to be evaluated
      @param env   an {@link Environment} for local variables
      @return the {@link Value} of the expression {@link Tree}
   */
   public static Value evaluateProg(final Tree<Token> tree,
                                    final Environment env)
   {
      final Value result;  // a blank final

      // Check which kind of Prog we have.
      if ( ! tree.root().lexeme.equals("prog") )
      {
         // Evaluate the single expression.
         result = evaluateExpr( tree, env );
      }
      else
      {
         // Evaluate each Expr in the Prog.
         // Any Var expressions will have the side effect
         // of putting a variable in the environment.
         // Any Set expressions will have the side effect
         // of changing a value in the environment.
         // Any Print expressions will have the side effect
         // of printing an output.
         // Any other expressions would be pointless!
         for (int i = 0; i < tree.subTrees().size()-1; ++i)
         {
            evaluateExpr( tree.subTrees().get(i), env );
         }

         // Evaluate the last expression and use its
         // value as the value of the prog expression.
         result = evaluateExpr( tree.subTrees().get(tree.subTrees().size()-1), env );
      }

      return result;
   }


   // Evaluate an expression.
   private static Value evaluateExpr(final Tree<Token> tree,
                                     final Environment env)
   {
      final Value result;  // a blank final

      final String node = tree.root().lexeme;

      if ( node.equals("begin") )
      {
         result = evaluateBegin(tree, env);
      }
      else if ( node.equals("set") )
      {
         result = evaluateSet(tree, env);
      }
      else if ( node.equals("var") )
      {
         result = evaluateVar(tree, env);
      }
      else if ( node.equals("print") )
      {
         result = evaluatePrint(tree, env);
      }
      else if ( node.equals("&&")
             || node.equals("||")
             || node.equals("!") )
      {
         result = evaluateBexpr(tree, env);  // boolean expression
      }
      else if ( node.equals("==")
             || node.equals("!=") )
      {
         result = evaluateEqexpr(tree, env);  // equality operator
      }
      else if ( node.equals("<")
             || node.equals(">")
             || node.equals("<=")
             || node.equals(">=") )
      {
         result = evaluateRelexpr(tree, env);  // relational operator
      }
      else if ( node.equals("+")
             || node.equals("-")
             || node.equals("*")
             || node.equals("/")
             || node.equals("%")
             || node.equals("**") )
      {
         result = evaluateAexpr(tree, env);  // arithmetic expression
      }
      else if ( tree.subTrees().size() == 0 ) // a leaf node
      {
         if ( node.equals("true") || node.equals("false") )
         {
            result = new Value( node.equals("true") );
         }
         else if ( node.matches("^[-|+]*[0-9][0-9]*") )
         {
            result = new Value( Integer.parseInt( node ) );
         }
         else if ( env.defined(node) )  // a variable
         {
            result = env.lookUp( node );
         }
         else
         {
            throw new EvalError("undefined variable: " + node);
         }
      }
      else
      {
         throw new EvalError("invalid expression: " + toString(tree));
      }

      return result;
   }


   // Evaluate a begin expression.
   private static Value evaluateBegin(final Tree<Token> tree,
                                      final Environment env)
   {
      // Create a new Environment object chained to (or "nested in")
      // the previous ("outer") environment object.
      final Environment newEnv = new Environment(env);

      // Evaluate each sub expression in the begin
      // expression (using the new environment chain).
      // The return value of each expression is
      // discarded, so any expression without a
      // side-effect is worthless.
      for (int i = 0; i < tree.subTrees().size()-1; ++i)
      {
         evaluateExpr( tree.subTrees().get(i), newEnv );
      }

      // Evaluate the last expression and use its
      // value as the value of the begin expression.
      final Value result =
         evaluateExpr(tree.subTrees().get(tree.subTrees().size()-1), newEnv);

      // When this method returns, the local Environment
      // object that we created at the beginning of this method
      // becomes a garbage object (and will be garbage collected).
      // This is like "popping" a stack frame off of the call stack
      // in Java, C , or C++.

      return result;
   }


   // Evaluate a set expression.
   private static Value evaluateSet(final Tree<Token> tree,
                                    final Environment env)
   {
      // get the variable
      final String variable = tree.subTrees().get(0).root().lexeme;

      // check if this variable has already been declared
      if ( ! env.defined(variable) )
      {
         throw new EvalError("undefined variable: " + variable);
      }

      // get, and then evaluate, the expression
      final Tree<Token> expr = tree.subTrees().get(1);
      final Value result = evaluateExpr( expr, env );
      // update this variable in the environment
      env.update(variable, result);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }


   // Evaluate a var expression.
   private static Value evaluateVar(final Tree<Token> tree,
                                    final Environment env)
   {
      // get the variable
      final String variable = tree.subTrees().get(0).root().lexeme;

      // check if this variable has already been declared
      // in the local environment
      if ( env.definedLocal(variable) )
      {
         throw new EvalError("variable already declared: " + variable);
      }

      // get, and then evaluate, the expression
      final Tree<Token> expr = tree.subTrees().get(1);
      final Value result = evaluateExpr( expr, env );

      // declare the new, local, variable
      env.add(variable, result);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }


   // Evaluate a print expression.
   private static Value evaluatePrint(final Tree<Token> tree,
                                      final Environment env)
   {
      final Value result = evaluateExpr( tree.subTrees().get(0), env );

      if (DEBUG > 0)
      {
         System.out.println( result );
      }
      else
      {
         System.out.println( result.toSimpleString() );
      }

      return result;
   }


   // Evaluate a boolean expression.
   private static Value evaluateBexpr(final Tree<Token> tree,
                                      final Environment env)
   {
      boolean result = false;

      final String node = tree.root().lexeme;

      Value value = evaluateExpr( tree.subTrees().get(0), env );

      result = value.valueB;

      if ( node.equals("&&") )
      {
         for (int i = 1; i < tree.subTrees().size(); ++i)
         {
            if (result)
            {
               value = evaluateExpr( tree.subTrees().get(i), env );

               result = result && value.valueB;
            }
            else  // short circuit the evaluation of '&&'
            {
               result = false;
               break;
            }
         }
      }
      else if ( node.equals("||") )
      {
         for (int i = 1; i < tree.subTrees().size(); ++i)
         {
            if (! result)
            {
               value = evaluateExpr( tree.subTrees().get(i), env );

               result = result || value.valueB;
            }
            else  // short circuit the evaluation of '||'
            {
               result = true;
               break;
            }
         }
      }
      else if ( node.equals("!") )
      {
         result = ! result;
      }

      return new Value( result );
   }


   // Evaluate an equality expression (which is a boolean expression).
   private static Value evaluateEqexpr(final Tree<Token> tree,
                                       final Environment env)
   {
      boolean result = false;

      final String op = tree.root().lexeme;

      final Value valueL = evaluateExpr( tree.subTrees().get(0), env );
      final Value valueR = evaluateExpr( tree.subTrees().get(1), env );

      if ( op.equals("==") )
      {
         if ( ! valueL.tag.equals(valueR.tag) )
         {
            result = false;
         }
         else if ( valueL.tag.equals(Value.INT_TAG) )
         {
            final int resultL = valueL.valueI;
            final int resultR = valueR.valueI;
            result = resultL == resultR;
         }
         else // boolean data type
         {
            final boolean resultL = valueL.valueB;
            final boolean resultR = valueR.valueB;
            result = resultL == resultR;
         }
      }
      else // the '!=' operator
      {
         if ( ! valueL.tag.equals(valueR.tag) )
         {
            result = true;
         }
         else if ( valueL.tag.equals(Value.INT_TAG) )
         {
            final int resultL = valueL.valueI;
            final int resultR = valueR.valueI;
            result = resultL != resultR;
         }
         else // boolean data type
         {
            final boolean resultL = valueL.valueB;
            final boolean resultR = valueR.valueB;
            result = resultL != resultR;
         }
      }

      return new Value( result );
   }


   // Evaluate a relational expression (which is a boolean expression).
   private static Value evaluateRelexpr(final Tree<Token> tree,
                                        final Environment env)
   {
      boolean result = false;

      final String op = tree.root().lexeme;

      Value valueL = evaluateExpr( tree.subTrees().get(0), env );

      Value valueR = evaluateExpr( tree.subTrees().get(1), env );

      final int resultL = valueL.valueI;
      final int resultR = valueR.valueI;

      if ( op.equals("<") )
      {
         result = resultL < resultR;
      }
      else if ( op.equals(">") )
      {
         result = resultL > resultR;
      }
      else if ( op.equals("<=") )
      {
         result = resultL <= resultR;
      }
      else if ( op.equals(">=") )
      {
         result = resultL >= resultR;
      }

      return new Value( result );
   }


   // Evaluate an arithmetic expression.
   private static Value evaluateAexpr(final Tree<Token> tree,
                                      final Environment env)
   {
      int result = 0;

      final String node = tree.root().lexeme;

      final Value valueL = evaluateExpr( tree.subTrees().get(0), env );

      final int resultL = valueL.valueI;
            int resultR = 0;

      if ( tree.subTrees().size() >= 2 )
      {
         final Value valueR = evaluateExpr( tree.subTrees().get(1), env );

         resultR = valueR.valueI;
      }

      if ( node.equals("+") )
      {
         if ( tree.subTrees().size() == 1 )
            result = resultL;
         else
         {
            result = resultL + resultR;

            for (int i = 2; i < tree.subTrees().size(); ++i)
            {
               final Value temp = evaluateExpr( tree.subTrees().get(i), env );

               result += temp.valueI;
            }
         }
      }
      else if ( node.equals("-") )
      {
         if ( tree.subTrees().size() == 1 )
            result = -resultL;
         else
            result = resultL - resultR;
      }
      else if ( node.equals("*") )
      {
         result = resultL * resultR;

         for (int i = 2; i < tree.subTrees().size(); ++i)
         {
            final Value temp = evaluateExpr( tree.subTrees().get(i), env );

            result *= temp.valueI;
         }
      }
      else if ( node.equals("/") )
      {
         result = resultL / resultR;
      }
      else if ( node.equals("%") )
      {
         result = resultL % resultR;
      }
      else if ( node.equals("**") )
      {
         result = (int)Math.pow(resultL, resultR);
      }

      return new Value( result );
   }


   /**
      Convert a {@code Tree} of {@link Token}s into a {@link String}.
      <p>
      This method does a pre-order traversal of the {@link Tree}.
      <p>
      This method is used by the evaluation errors.

      @param tree  {@link Tree} of {@link Token}s to convert into a {@link String}
      @return a {@link String} representation of this {@code Tree} of {@link Token}s
   */
   public static String toString(final Tree<Token> tree)
   {
      String result = "";
      if ( tree.isLeaf() )
      {
         result += tree.root().lexeme;
      }
      else
      {
         result += "(" + tree.root().lexeme;
         for ( final Tree<Token> subTree : tree.subTrees() )
         {
            result += " " + toString(subTree);  // recursion
         }
         result += ")";
      }
      return result;
   }
}
